@@ -1,7 +1,7 @@
-'''pydons = Python numerical data manimulation add-ons
+'''Pydons is a collection of manipulation add-ons for hierarchichal numerical data.
 '''
 
-__version__ = '0.1.1'
+__version__ = '0.2.0'
 
 # import OrderedDict for Python < 2.7
 # In Python 2.6 or Jython 2.5, ordereddict must be installed
@@ -20,6 +20,11 @@ import numbers
 import pydons.hdf5util
 import hdf5storage
 import numpy as np
+import posixpath
+import netCDF4
+import h5py
+import os
+import sys
 
 
 class MatStruct(_OrderedDict):
@@ -35,7 +40,7 @@ class MatStruct(_OrderedDict):
 
     __FORBIDDEN_KEYS = tuple(dir(_OrderedDict) +
                              ['insert_after', 'insert_before',
-                              'diff', 'saveh5', 'loadh5',
+                              'diff', 'merge', 'saveh5', 'loadh5',
                               'savemat', 'loadmat'])
     __MC = None
 
@@ -65,6 +70,8 @@ class MatStruct(_OrderedDict):
         return True
 
     def __init__(self, *args, **kwargs):
+        # hiding attributes via __dir__ does not seem to work in ipython
+        # self._hide_methods = kwargs.pop('hide_methods', False)
         super(MatStruct, self).__init__(*args, **kwargs)
 
     def __getattr__(self, item):
@@ -123,6 +130,7 @@ class MatStruct(_OrderedDict):
     def __dir__(self):
         d = []
         d += self.keys()
+        # if not self._hide_methods:
         d += dir(self.__class__)
         d += self.__dict__.keys()
         d.sort()
@@ -229,6 +237,14 @@ class MatStruct(_OrderedDict):
         res['diff_norm'] /= nnorm
         return res
 
+    def merge(self, other):
+        '''Merge fields form another MatStruct or any dict-like object
+
+        :param other: object to merge from
+        '''
+        # TODO
+        raise NotImplementedError('to be implemented')
+
     def saveh5(self, file_name, path='/', truncate_existing=False,
                matlab_compatible=False, **kwargs):
         """Save to an HDF5 file
@@ -258,6 +274,7 @@ class MatStruct(_OrderedDict):
         :param file_name: output file name
         :param path: group path to store fields to
         """
+        # TODO switch convert ints to doubles
         self.saveh5(file_name, path, truncate_existing=truncate_existing,
                     matlab_compatible=True, **kwargs)
 
@@ -269,3 +286,163 @@ class MatStruct(_OrderedDict):
         :param path: path toread data from
         """
         return cls.loadh5(file_name, path, matlab_compatible=True, **kwargs)
+
+
+class NC4File(netCDF4.Dataset):
+    """NetCDF 4 file with __getitem__"""
+    def __init__(self, *args, **argv):
+        super(NC4File, self).__init__(*args, **argv)
+
+    def __getitem__(self, key):
+        '''Get item from a key specified as a posix path'''
+        grp = self
+        # remove leading /
+        while key.startswith('/'):
+            key = key[1:]
+        if not key:
+            # get the root
+            return self
+        key = posixpath.normpath(key)
+        keys = key.split('/')
+        grps, var = keys[:-1], keys[-1]
+        # get the final group
+        for k in grps:
+            grp = grp.groups[k]
+        # get the variable or group
+        if var in grp.variables:
+            return grp.variables[var]
+        elif var in grp.groups:
+            return grp.groups[var]
+        else:
+            raise KeyError('%s not found' % key)
+
+
+class LazyDataset(object):
+    """NetCDF 4 / HDF5 data set object with lazy evaluation"""
+    def __init__(self, grp, name, lazy_min_size=100, lazy_max_size=100000000):
+        if isinstance(grp, (netCDF4.Group, netCDF4.Dataset)):
+            self._fileclass = NC4File
+            self._filepath = os.path.abspath(grp.filepath())
+            self._path = posixpath.join(grp.path, name)
+            dset = grp.variables[name]
+        elif isinstance(grp, h5py.Group):
+            self._fileclass = h5py.File
+            self._filepath = os.path.abspath(grp.file.filename)
+            self._path = posixpath.join(grp.name, name)
+            self._data = None
+            dset = grp[name]
+        else:
+            raise TypeError('%s not supported' % type(grp))
+        self._data = None
+        self._lazy_min_size = lazy_min_size
+        self._lazy_max_size = lazy_max_size
+        # preloaded attributes
+        for prop in ('dtype', 'shape', 'size', 'ndim', 'dimensions', 'title', 'units'):
+            if hasattr(dset, prop):
+                setattr(self, prop, getattr(dset, prop))
+        if self.size <= lazy_min_size:
+            self._get_data()
+
+    def _get_data(self):
+        if self._data is None:
+            with self._fileclass(self._filepath, 'r') as f:
+                data = f[self._path][:]
+                if data.size <= self._lazy_max_size:
+                    self._data = data
+            return data
+        else:
+            return self._data
+
+    def __getitem__(self, key):
+        """Get slice (read rada)"""
+        return self._get_data()[key]
+
+    def __getattr__(self, attr):
+        if self._data:
+            return getattr(self._data, attr)
+        else:
+            raise AttributeError('no attribute %s' % attr)
+            # return getattr(self._fileclass(self._filepath, 'r')[self._path], attr)
+
+    def __iter__(self):
+        # iterate over data
+        return iter(self._get_data())
+
+    def __array__(self):
+        # for numpy
+        return self._get_data()
+
+    if sys.version_info < (2, 0):
+        # They won't be defined if version is at least 2.0 final
+        def __getslice__(self, i, j):
+            return self[max(0, i):max(0, j):]
+
+
+class FileBrowser(MatStruct):
+    """Load netCDF of HDF5 file into an offline MatStruct"""
+    def __init__(self, file_name, lazy_min_size=100, lazy_max_size=100000000):
+        """Read hierarchical data file into a MatStruct tree with data in LazyDataset
+
+        :param file_name: file name
+        :param lazy_min_size: data sets with a lower size will be always stored in the memory
+        :param lazy_max_size: data sets with a larger size will never be stored in the memory
+        """
+        super(FileBrowser, self).__init__()
+        # get the file type and corresponding classes
+        _, ext = os.path.splitext(file_name)
+        ext = ext[1:]
+        if ext.lower() in ('nc', 'cdf'):
+            fileclass, dataclass = NC4File, LazyDataset
+        elif ext.lower() in ('h5', 'hdf5', 'he5'):
+            fileclass, dataclass = h5py.File, LazyDataset
+        else:
+            raise TypeError('Unknow file extension: %s' % ext)
+        # recursively read the file structure
+        with fileclass(file_name, 'r') as fileobj:
+            for key, val in _read_all(fileobj, dataclass,
+                                      lazy_min_size=lazy_min_size,
+                                      lazy_max_size=lazy_max_size).items():
+                self[key] = val
+
+
+def _read_all(basegrp, dataclass, lazy_min_size, lazy_max_size):
+    """Recursively read all groups / variables
+
+    :param basegrp: base group (the starting point)
+    :param dataclass: data type to return
+    """
+    res = MatStruct()
+    for grpname in _groups(basegrp):
+        try:
+            res[grpname] = _read_all(_groups(basegrp)[grpname], dataclass,
+                                     lazy_min_size, lazy_max_size)
+        except KeyError:
+            res[grpname + '_'] = _read_all(_groups(basegrp)[grpname], dataclass,
+                                           lazy_min_size, lazy_max_size)
+    for varname in _variables(basegrp):
+        try:
+            res[varname] = dataclass(basegrp, varname, lazy_min_size, lazy_max_size)
+        except KeyError:
+            res[varname + '_'] = dataclass(basegrp, varname, lazy_min_size, lazy_max_size)
+    return res
+
+
+def _groups(basegrp):
+    """Get group names from an HDF5/netCDF4 file group
+    """
+    if hasattr(basegrp, 'groups'):
+        return basegrp.groups
+    else:
+        # for HDF5 return keys that are dict-like
+        # return [key for key in basegrp.keys() if hasattr(basegrp[key], 'keys')]
+        return _OrderedDict([(key, value) for key, value in basegrp.items() if hasattr(value, 'keys')])
+
+
+def _variables(basegrp):
+    """Get group names from an HDF5/netCDF4 file group
+    """
+    if hasattr(basegrp, 'variables'):
+        return basegrp.variables
+    else:
+        # for HDF5 return keys that are not dict-like
+        return _OrderedDict([(key, value) for key, value in basegrp.items() if not hasattr(value, 'keys')])
